/**
 *
 * SIROCCO
 * Copyright (C) 2011 France Telecom
 * Contact: sirocco@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  $Id: VMwareCustomizationSpecBuilder.java 13 2011-06-14 21:21:51Z dangtran $
 *
 */

package org.ow2.sirocco.vmm.agent.driver.vmware;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Properties;

import org.apache.log4j.Logger;
import org.ow2.sirocco.vmm.agent.main.AgentControl;
import org.ow2.sirocco.vmm.api.VnicIPSettings;
import org.ow2.sirocco.vmm.api.VnicIPSettings.IpAssignmentMode;

import com.vmware.vim.CustomizationAdapterMapping;
import com.vmware.vim.CustomizationDhcpIpGenerator;
import com.vmware.vim.CustomizationFixedIp;
import com.vmware.vim.CustomizationFixedName;
import com.vmware.vim.CustomizationGlobalIPSettings;
import com.vmware.vim.CustomizationGuiUnattended;
import com.vmware.vim.CustomizationIPSettings;
import com.vmware.vim.CustomizationIdentification;
import com.vmware.vim.CustomizationLicenseDataMode;
import com.vmware.vim.CustomizationLicenseFilePrintData;
import com.vmware.vim.CustomizationLinuxPrep;
import com.vmware.vim.CustomizationPassword;
import com.vmware.vim.CustomizationSpec;
import com.vmware.vim.CustomizationSysprep;
import com.vmware.vim.CustomizationUserData;
import com.vmware.vim.CustomizationWinOptions;

public class VMwareCustomizationSpecBuilder {
    static Logger logger = Logger.getLogger(VMwareCustomizationSpecBuilder.class);

    static private String WINDOWS_SYSPREP_PASSWORD_KEY = "windows.sysprep.password";

    static private String WINDOWS_SYSPREP_LICENSE_TYPE_KEY = "windows.sysprep.license_type";

    static private String WINDOWS_SYSPREP_LICENSE_TYPE_PER_SEAT_VALUE = "per_seat";

    static private String WINDOWS_SYSPREP_LICENSE_TYPE_PER_SERVER_VALUE = "per_server";

    static private String WINDOWS_SYSPREP_WORKGROUP_KEY = "windows.sysprep.workgroup";

    static private String WINDOWS_SYSPREP_AUTO_LOGON_COUNT_KEY = "windows.sysprep.auto_logon_count";

    static private String WINDOWS_SYSPREP_TIMEZONE_KEY = "windows.sysprep.timezone";

    static private String WINDOWS_SYSPREP_USERDATA_FULLNAME_KEY = "windows.sysprep.userdata_fullname";

    static private String WINDOWS_SYSPREP_USERDATA_ORGNAME_KEY = "windows.sysprep.userdata_orgname";

    static private String WINDOWS_SYSPREP_USERDATA_PRODUCTID_KEY = "windows.sysprep.userdata_product_key";

    static private Properties defaultSysprepProperties = new Properties();

    static {
        String winGuestPropFileName = "etc" + File.separator + "windows-sysprep.properties";
        try {
            VMwareCustomizationSpecBuilder.defaultSysprepProperties.load(new FileInputStream(new File(AgentControl.vmmHome,
                winGuestPropFileName)));
        } catch (IOException ex) {
            VMwareCustomizationSpecBuilder.logger.warn("Cannot load Windows sysprep property file " + winGuestPropFileName);
        }
    }

    private static String getWinProp(final org.ow2.sirocco.vmm.api.CustomizationSpec vmSpec, final String name) {
        String val = vmSpec.getGuestOsCustomizationParams().get(name);
        if (val == null || val.length() == 0) {
            val = VMwareCustomizationSpecBuilder.defaultSysprepProperties.getProperty(name);
        }
        return val;
    }

    /**
     * Prepares a CustomizationSpec object to be used when cloning a VM. Most
     * supported customization parameters except network-related ones are used
     * to sysprep Windows OS
     * 
     * @param vmSpec VMConfigSpec object containing customization parameters
     * @param isWindows true if the VM to customize contains a Windows guest OS
     * @return a VMWare API CustomizationSpec object
     */
    public static CustomizationSpec buildCustomizationSpec(final org.ow2.sirocco.vmm.api.CustomizationSpec vmSpec,
        final boolean isWindows) {
        CustomizationSpec custSpec = new CustomizationSpec();

        if (isWindows) {
            CustomizationSysprep custSysprep = new CustomizationSysprep();

            String password = VMwareCustomizationSpecBuilder.getWinProp(vmSpec,
                VMwareCustomizationSpecBuilder.WINDOWS_SYSPREP_PASSWORD_KEY);
            String autoLogonCount = VMwareCustomizationSpecBuilder.getWinProp(vmSpec,
                VMwareCustomizationSpecBuilder.WINDOWS_SYSPREP_AUTO_LOGON_COUNT_KEY);
            String timeZone = VMwareCustomizationSpecBuilder.getWinProp(vmSpec,
                VMwareCustomizationSpecBuilder.WINDOWS_SYSPREP_TIMEZONE_KEY);
            if (autoLogonCount != null && timeZone != null) {
                VMwareCustomizationSpecBuilder.logger.debug("Customizing VM : password=" + password + " autoLogonCount="
                    + autoLogonCount + " timeZone=" + timeZone);
                CustomizationGuiUnattended custGuiUnattended = new CustomizationGuiUnattended();
                if (password != null) {
                    CustomizationPassword custPassword = new CustomizationPassword();
                    custPassword.setPlainText(true);
                    custPassword.setValue(password);
                    custGuiUnattended.setPassword(custPassword);
                }
                custGuiUnattended.setAutoLogon(true);
                custGuiUnattended.setAutoLogonCount(Integer.parseInt(autoLogonCount));
                custGuiUnattended.setTimeZone(Integer.parseInt(timeZone));
                custSysprep.setGuiUnattended(custGuiUnattended);
            }

            String licenseType = VMwareCustomizationSpecBuilder.getWinProp(vmSpec,
                VMwareCustomizationSpecBuilder.WINDOWS_SYSPREP_LICENSE_TYPE_KEY);
            if (licenseType != null) {
                if (licenseType.equalsIgnoreCase(VMwareCustomizationSpecBuilder.WINDOWS_SYSPREP_LICENSE_TYPE_PER_SEAT_VALUE)) {
                    CustomizationLicenseFilePrintData custLPD = new CustomizationLicenseFilePrintData();
                    custLPD.setAutoMode(CustomizationLicenseDataMode.perSeat);
                    custSysprep.setLicenseFilePrintData(custLPD);
                } else if (licenseType
                    .equalsIgnoreCase(VMwareCustomizationSpecBuilder.WINDOWS_SYSPREP_LICENSE_TYPE_PER_SERVER_VALUE)) {
                    CustomizationLicenseFilePrintData custLPD = new CustomizationLicenseFilePrintData();
                    custLPD.setAutoMode(CustomizationLicenseDataMode.perServer);
                    custSysprep.setLicenseFilePrintData(custLPD);
                } else {
                    VMwareCustomizationSpecBuilder.logger.warn("unknown license type: " + licenseType);
                }
            }

            String workgroup = VMwareCustomizationSpecBuilder.getWinProp(vmSpec,
                VMwareCustomizationSpecBuilder.WINDOWS_SYSPREP_WORKGROUP_KEY);
            if (workgroup != null) {
                CustomizationIdentification custId = new CustomizationIdentification();
                custId.setJoinWorkgroup(workgroup);
                custSysprep.setIdentification(custId);
            }

            String fullName = VMwareCustomizationSpecBuilder.getWinProp(vmSpec,
                VMwareCustomizationSpecBuilder.WINDOWS_SYSPREP_USERDATA_FULLNAME_KEY);
            String orgName = VMwareCustomizationSpecBuilder.getWinProp(vmSpec,
                VMwareCustomizationSpecBuilder.WINDOWS_SYSPREP_USERDATA_ORGNAME_KEY);
            String productId = VMwareCustomizationSpecBuilder.getWinProp(vmSpec,
                VMwareCustomizationSpecBuilder.WINDOWS_SYSPREP_USERDATA_PRODUCTID_KEY);

            if (fullName != null && orgName != null && productId != null && vmSpec.getGuestOsHostName() != null) {
                CustomizationUserData custUserData = new CustomizationUserData();
                custUserData.setFullName(fullName);
                custUserData.setOrgName(orgName);
                custUserData.setProductId(productId);
                VMwareCustomizationSpecBuilder.logger.debug("Customizing VM : productId=" + productId);
                CustomizationFixedName custFixedName = new CustomizationFixedName();
                VMwareCustomizationSpecBuilder.logger.debug("Customizing VM : hostname=" + vmSpec.getGuestOsHostName());
                custFixedName.setName(vmSpec.getGuestOsHostName());
                custUserData.setComputerName(custFixedName);
                custSysprep.setUserData(custUserData);
            }

            custSpec.setIdentity(custSysprep);

            CustomizationWinOptions custWinOptions = new CustomizationWinOptions();
            custWinOptions.setChangeSID(true);
            custSpec.setOptions(custWinOptions);
        } else {
            // Linux guest
            CustomizationLinuxPrep custLinuxPrep = new CustomizationLinuxPrep();
            if (vmSpec.getGuestOsDomain() == null) {
                custLinuxPrep.setDomain("");
            } else {
                custLinuxPrep.setDomain(vmSpec.getGuestOsDomain());
            }
            CustomizationFixedName fixedName = new CustomizationFixedName();
            fixedName.setName(vmSpec.getGuestOsHostName());
            custLinuxPrep.setHostName(fixedName);
            custSpec.setIdentity(custLinuxPrep);
        }

        CustomizationGlobalIPSettings custGlobalIPSettings = new CustomizationGlobalIPSettings();
        custSpec.setGlobalIPSettings(custGlobalIPSettings);

        // TODO support for multiple VNICs
        if (vmSpec.getVnicIpSettingsList() != null && !vmSpec.getVnicIpSettingsList().isEmpty()) {
            custGlobalIPSettings.setDnsServerList(vmSpec.getGuestOsDnsServerList().toArray(new String[0]));
            custGlobalIPSettings.setDnsSuffixList(new String[] {vmSpec.getGuestOsDomain()});

            VnicIPSettings vnicIPSettings = vmSpec.getVnicIpSettingsList().get(0);

            CustomizationAdapterMapping custAdapterMapping = new CustomizationAdapterMapping();
            CustomizationIPSettings custIPSettings = new CustomizationIPSettings();
            custIPSettings.setDnsDomain(vmSpec.getGuestOsDomain());
            VMwareCustomizationSpecBuilder.logger.debug("Customizing VM : DNS server="
                + vmSpec.getGuestOsDnsServerList().get(0));
            custIPSettings.setDnsServerList(vmSpec.getGuestOsDnsServerList().toArray(new String[0]));

            VMwareCustomizationSpecBuilder.logger.debug("Customizing VM : gateway=" + vnicIPSettings.getGateway());
            custIPSettings.setGateway(new String[] {vnicIPSettings.getGateway()});
            custIPSettings.setSubnetMask(vnicIPSettings.getSubnetMask());
            if (vnicIPSettings.getIpAssignmentMode() == IpAssignmentMode.FIXED) {
                CustomizationFixedIp custFixedIp = new CustomizationFixedIp();
                VMwareCustomizationSpecBuilder.logger.debug("Customizing VM: VNIC IP address=" + vnicIPSettings.getIpAddress());
                custFixedIp.setIpAddress(vnicIPSettings.getIpAddress());
                custIPSettings.setIp(custFixedIp);
            } else if (vnicIPSettings.getIpAssignmentMode() == IpAssignmentMode.DHCP) {
                CustomizationDhcpIpGenerator custDhcp = new CustomizationDhcpIpGenerator();
                custIPSettings.setIp(custDhcp);
            }
            custAdapterMapping.setAdapter(custIPSettings);

            custSpec.setNicSettingMap(new CustomizationAdapterMapping[] {custAdapterMapping});
        } else {
            CustomizationAdapterMapping custAdapterMapping = new CustomizationAdapterMapping();
            CustomizationIPSettings custIPSettings = new CustomizationIPSettings();
            CustomizationDhcpIpGenerator custDhcp = new CustomizationDhcpIpGenerator();
            custIPSettings.setIp(custDhcp);
            custAdapterMapping.setAdapter(custIPSettings);
            custSpec.setNicSettingMap(new CustomizationAdapterMapping[] {custAdapterMapping});
        }

        return custSpec;
    }
}
