/**
 *
 * SIROCCO
 * Copyright (C) 2011 France Telecom
 * Contact: sirocco@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  $Id: VMwareStoragePool.java 392 2011-10-03 13:37:36Z dangtran $
 *
 */

package org.ow2.sirocco.vmm.agent.driver.vmware;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.management.ObjectName;

import org.apache.log4j.Logger;
import org.ow2.sirocco.vmm.agent.domain.AbstractStoragePool;
import org.ow2.sirocco.vmm.api.InsufficientResourcesException;
import org.ow2.sirocco.vmm.api.StoragePoolMXBean;
import org.ow2.sirocco.vmm.api.VMMException;
import org.ow2.sirocco.vmm.api.Volume;

import com.vmware.vim.ArrayOfHostDatastoreBrowserSearchResults;
import com.vmware.vim.ArrayOfManagedObjectReference;
import com.vmware.vim.ArrayOfVirtualDevice;
import com.vmware.vim.FileBackedVirtualDiskSpec;
import com.vmware.vim.FileInfo;
import com.vmware.vim.FileQuery;
import com.vmware.vim.FileQueryFlags;
import com.vmware.vim.HostDatastoreBrowserSearchResults;
import com.vmware.vim.HostDatastoreBrowserSearchSpec;
import com.vmware.vim.IsoImageFileInfo;
import com.vmware.vim.IsoImageFileQuery;
import com.vmware.vim.ManagedObjectReference;
import com.vmware.vim.TaskInfo;
import com.vmware.vim.VirtualDevice;
import com.vmware.vim.VirtualDeviceFileBackingInfo;
import com.vmware.vim.VirtualDisk;
import com.vmware.vim.VirtualDiskAdapterType;
import com.vmware.vim.VirtualDiskType;
import com.vmware.vim.VmDiskFileInfo;
import com.vmware.vim.VmDiskFileQuery;
import com.vmware.vim.VmDiskFileQueryFlags;

public class VMwareStoragePool extends AbstractStoragePool implements StoragePoolMXBean {
    static Logger logger = Logger.getLogger(VMwareStoragePool.class);

    static final int TEMPLATE_DISK = 1;

    static final int VM_DISK = 2;

    static final int ISO_IMAGE = 4;

    static final String VMWARE_DISK_FORMAT = "vmdk";

    private String name;

    private VMwareServerPool serverPool;

    private VMwareServerPool.ConnectionPool vCenterConnectionPool;

    private VMwareServerPool.ConnectionPool hostConnectionPool;

    private String dataStoreName;

    private String filePath;

    private int flags;

    private ManagedObjectReference dataStoreRef;

    private ManagedObjectReference virtualDiskManagerRef;

    private Map<String, Volume> volumeMap = new HashMap<String, Volume>();

    private Map<String, ManagedObjectReference> templateMap = new HashMap<String, ManagedObjectReference>();

    private Map<String, String> properties;

    VMwareStoragePool(final String name, final ObjectName objectName, final VMwareServerPool serverPool,
        final String dataStoreName, final String filePath, final int flags, final Map<String, String> properties)
        throws Exception {
        super(objectName);
        this.name = name;
        this.serverPool = serverPool;
        this.vCenterConnectionPool = serverPool.getConnectionPool();
        this.dataStoreName = dataStoreName;
        this.filePath = filePath;
        this.flags = flags;
        this.properties = new HashMap<String, String>(properties);

        String host = properties.get("host");
        String login = properties.get("login");
        String password = properties.get("password");

        if (host != null) {
            if (login != null && password != null) {
                this.hostConnectionPool = new VMwareServerPool.ConnectionPool(serverPool.getExecutorService(), 1, host, login,
                    password);
            } else {
                VMwareStoragePool.logger.warn("Missing login and/or password for host " + host + " in storage pool " + name);
            }
        }

        this.initialize();
        if (this.dataStoreRef == null || this.virtualDiskManagerRef == null) {
            throw new Exception("Failed to init storage pool " + dataStoreName);
        }
        this.populate();
    }

    private void initialize() throws Exception {
        VMwareServiceConnection connection = this.vCenterConnectionPool.getConnection();
        try {
            ArrayOfManagedObjectReference dataStoreRefs = (ArrayOfManagedObjectReference) connection.getDynamicProperties(
                this.serverPool.getDatacenterRef(), "datastore")[0].getVal();
            if (dataStoreRefs != null) {
                for (int i = 0; i < dataStoreRefs.getManagedObjectReference().length; i++) {
                    String dsName = (String) connection.getDynamicProperties(dataStoreRefs.getManagedObjectReference(i),
                        "info.name")[0].getVal();
                    if (dsName.equals(this.dataStoreName)) {
                        this.dataStoreRef = dataStoreRefs.getManagedObjectReference(i);
                        break;
                    }
                }
            }
            this.virtualDiskManagerRef = connection.getServiceContent().getVirtualDiskManager();
        } finally {
            connection.release();
        }
        if (this.hostConnectionPool != null) {
            connection = this.hostConnectionPool.getConnection();
            try {
                this.virtualDiskManagerRef = connection.getServiceContent().getVirtualDiskManager();
            } finally {
                connection.release();
            }
        }
    }

    ManagedObjectReference getDataStore() {
        return this.dataStoreRef;
    }

    @Override
    public void release() {
        if (this.hostConnectionPool != null) {
            this.hostConnectionPool.stop();
            this.hostConnectionPool = null;
        }
    }

    @Override
    public void refresh() {
        try {
            this.populate();
        } catch (Exception ex) {
            VMwareStoragePool.logger.error("Failed to refresh datastore " + this.dataStoreName, ex);
        }
    }

    private synchronized void populate() throws Exception {
        this.volumeMap.clear();
        this.templateMap.clear();
        VMwareServiceConnection connection = this.vCenterConnectionPool.getConnection();
        try {
            this.scan(connection);
            if ((this.flags & VMwareStoragePool.TEMPLATE_DISK) != 0) {
                this.findTemplateVMDisks(connection);
            }
        } finally {
            connection.release();
        }
    }

    private void scan(final VMwareServiceConnection connection) throws Exception {
        ManagedObjectReference browserRef = (ManagedObjectReference) connection.getDynamicProperties(this.dataStoreRef,
            "browser")[0].getVal();
        HostDatastoreBrowserSearchSpec searchSpec = new HostDatastoreBrowserSearchSpec();
        int fileQueriesNum = 0;
        if ((this.flags & VMwareStoragePool.VM_DISK) != 0) {
            fileQueriesNum++;
        }
        if ((this.flags & VMwareStoragePool.ISO_IMAGE) != 0) {
            fileQueriesNum++;
        }
        FileQuery[] fileQueries = new FileQuery[fileQueriesNum];
        int i = 0;
        if ((this.flags & VMwareStoragePool.ISO_IMAGE) != 0) {
            fileQueries[i++] = new IsoImageFileQuery();
        }
        if ((this.flags & VMwareStoragePool.VM_DISK) != 0) {
            VmDiskFileQuery fileQuery = new VmDiskFileQuery();
            VmDiskFileQueryFlags flags = new VmDiskFileQueryFlags();
            flags.setCapacityKb(true);
            fileQuery.setDetails(flags);
            fileQueries[i++] = fileQuery;
        }
        searchSpec.setQuery(fileQueries);
        FileQueryFlags fileQueryFlags = new FileQueryFlags();
        fileQueryFlags.setFileSize(true);
        fileQueryFlags.setFileType(true);
        fileQueryFlags.setFileOwner(false);
        searchSpec.setDetails(fileQueryFlags);
        searchSpec.setQuery(fileQueries);

        ManagedObjectReference task = connection.getService().searchDatastoreSubFolders_Task(browserRef,
            this.getDataStorePath(), searchSpec);
        String status = connection.waitForTask(task);
        if (!status.equalsIgnoreCase("success")) {
            VMwareStoragePool.logger.error("Cannot search datastore " + this.dataStoreName + " path=" + this.filePath);
        } else {
            ArrayOfHostDatastoreBrowserSearchResults results = (ArrayOfHostDatastoreBrowserSearchResults) connection
                .getDynamicProperties(task, "info.result")[0].getVal();
            for (HostDatastoreBrowserSearchResults searchResults : results.getHostDatastoreBrowserSearchResults()) {
                String folderPath = searchResults.getFolderPath();
                if (searchResults.getFile() != null) {
                    for (FileInfo fileInfo : searchResults.getFile()) {
                        Volume vol = new Volume();
                        vol.setName(fileInfo.getPath());
                        vol.setPath(folderPath + fileInfo.getPath());
                        vol.setStoragePool(this);
                        vol.setKey(vol.getPath());
                        vol.setName(fileInfo.getPath());
                        vol.setAllocationMB(fileInfo.getFileSize() / (1024 * 1024));
                        if (fileInfo instanceof VmDiskFileInfo) {
                            VmDiskFileInfo vmDiskInfo = (VmDiskFileInfo) fileInfo;
                            if (vmDiskInfo.getCapacityKb() == null) {
                                vol.setCapacityMB(-1);
                            } else {
                                vol.setCapacityMB(vmDiskInfo.getCapacityKb() / 1024);
                            }
                            vol.setFormat(VMwareStoragePool.VMWARE_DISK_FORMAT);
                        } else if (fileInfo instanceof IsoImageFileInfo) {
                            IsoImageFileInfo isoImageInfo = (IsoImageFileInfo) fileInfo;
                            vol.setFormat("iso");
                        }
                        VMwareStoragePool.logger.debug("Added volume " + vol);
                        this.volumeMap.put(vol.getKey(), vol);
                    }
                }
            }
        }
    }

    private void findTemplateVMDisks(final VMwareServiceConnection connection) throws Exception {
        ArrayOfManagedObjectReference vmRefs = (ArrayOfManagedObjectReference) connection.getDynamicProperties(
            this.dataStoreRef, "vm")[0].getVal();
        if (vmRefs != null) {
            for (int i = 0; i < vmRefs.getManagedObjectReference().length; i++) {
                ManagedObjectReference vmRef = vmRefs.getManagedObjectReference(i);
                Boolean isTemplate = (Boolean) connection.getDynamicProperties(vmRef, "config.template")[0].getVal();
                if (isTemplate) {
                    this.addVMVolumes(connection, vmRef, true);
                }
            }
        }
    }

    List<Volume> addVMVolumes(final VMwareServiceConnection connection, final ManagedObjectReference vm,
        final boolean isTemplate) throws Exception {
        List<Volume> result = new ArrayList<Volume>();
        ArrayOfVirtualDevice devices = (ArrayOfVirtualDevice) connection.getDynamicProperties(vm, "config.hardware.device")[0]
            .getVal();
        for (VirtualDevice device : devices.getVirtualDevice()) {
            if (device instanceof VirtualDisk) {
                VirtualDisk disk = (VirtualDisk) device;
                VirtualDeviceFileBackingInfo fileBacking = (VirtualDeviceFileBackingInfo) disk.getBacking();

                Volume vol = this.volumeMap.get(fileBacking.getFileName());
                if (vol == null) {
                    vol = new Volume();
                    vol.setPath(fileBacking.getFileName());
                    vol.setStoragePool(this);
                    vol.setKey(vol.getPath());
                    vol.setAllocationMB(-1);
                    int index = vol.getPath().lastIndexOf('/');
                    if (index == -1) {
                        index = 0;
                    }
                    vol.setName(vol.getPath().substring(index + 1));
                    vol.setCapacityMB(disk.getCapacityInKB() / 1024);
                    vol.setFormat(VMwareStoragePool.VMWARE_DISK_FORMAT);
                    VMwareStoragePool.logger.debug("Added volume " + vol);
                    this.volumeMap.put(vol.getKey(), vol);
                }
                if (isTemplate) {
                    VMwareStoragePool.logger.debug("Tying volume " + vol.getName() + " with VM template");
                    this.templateMap.put(vol.getKey(), vm);
                }
                result.add(vol);
            }
        }
        return result;
    }

    @Override
    public String getName() {
        return this.name;
    }

    @Override
    public String[] getSupportedDiskImageFormats() {
        return new String[] {VMwareStoragePool.VMWARE_DISK_FORMAT};
    }

    String getDataStoreName() {
        return "[" + this.dataStoreName + "]";
    }

    String getDataStorePath() {
        return "[" + this.dataStoreName + "]" + this.filePath;
    }

    ManagedObjectReference getTemplateRef(final Volume vol) {
        return this.templateMap.get(vol.getKey());
    }

    @Override
    public long getCapacityMB() throws VMMException {
        VMwareServiceConnection connection = this.vCenterConnectionPool.getConnection();
        try {
            long capacityInBytes = (Long) connection.getDynamicProperties(this.dataStoreRef, "summary.capacity")[0].getVal();
            return capacityInBytes / (1024 * 1024);
        } catch (Exception ex) {
            VMwareServerPool.logger.error("Failed to get datastore capacity", ex);
            throw new VMMException("Failed to get datastore capacity", ex);
        } finally {
            connection.release();
        }
    }

    @Override
    public long getFreeSpaceMB() throws VMMException {
        VMwareServiceConnection connection = this.vCenterConnectionPool.getConnection();
        try {
            long freeSpaceInBytes = (Long) connection.getDynamicProperties(this.dataStoreRef, "summary.freeSpace")[0].getVal();
            return freeSpaceInBytes / (1024 * 1024);
        } catch (Exception ex) {
            VMwareServerPool.logger.error("Failed to get datastore capacity", ex);
            throw new VMMException("Failed to get datastore capacity", ex);
        } finally {
            connection.release();
        }
    }

    @Override
    public String getType() {
        return "file-backed";
    }

    @Override
    public synchronized List<Volume> getVolumes() throws VMMException {
        return new ArrayList<Volume>(this.volumeMap.values());
    }

    @Override
    public Volume getVolumeByKey(final String key) {
        Volume vol = this.volumeMap.get(key);
        return vol;
    }

    @Override
    public void destroyVolume(final String key) throws VMMException {
        Volume vol = this.volumeMap.remove(key);
        if (vol == null) {
            throw new VMMException("Volume with key=" + key + " not found");
        }
        if (this.hostConnectionPool != null) {
            throw new VMMException("This operation requires a connection to ESX host");
        }
        VMwareServiceConnection connection = this.hostConnectionPool.getConnection();
        ManagedObjectReference task;
        String status;
        TaskInfo taskInfo = null;
        try {
            task = connection.getService().deleteVirtualDisk_Task(this.virtualDiskManagerRef, vol.getPath(), null);
            status = connection.waitForTask(task);
            if (!status.equalsIgnoreCase("success")) {
                taskInfo = (TaskInfo) connection.getDynamicProperties(task, "info")[0].getVal();
            }
        } catch (Exception ex) {
            VMwareStoragePool.logger.error("Failed to destroy virtual disk " + key, ex);
            throw new VMMException("Failed to destroy volume " + key + " : " + ex.getMessage());
        } finally {
            connection.release();
        }
        ManagedObjectReference templateVm = this.templateMap.get(key);
        if (templateVm != null) {
            VMwareServiceConnection vCenterConnection = this.vCenterConnectionPool.getConnection();
            try {
                try {
                    vCenterConnection.getService().destroy_Task(templateVm);
                } catch (Exception ex) {
                    VMwareStoragePool.logger.error("Failed to destroy template VM tied with volume " + key, ex);
                    throw new VMMException("Failed to destroy volume " + key + " : " + ex.getMessage());
                }
            } finally {
                vCenterConnection.release();
            }
        }
        if (!status.equalsIgnoreCase("success")) {
            throw new VMMException("Failed to destroy volume : " + taskInfo.getError().getLocalizedMessage());
        }

    }

    @Override
    public Volume createVolume(final String name, final long capacityMB, final String format) throws VMMException {
        if (!format.equals(VMwareStoragePool.VMWARE_DISK_FORMAT)) {
            throw new VMMException("Unsupported image format: " + format);
        }
        if (this.hostConnectionPool != null) {
            throw new VMMException("This operation requires a connection to ESX host");
        }
        VMwareServiceConnection connection = this.hostConnectionPool.getConnection();
        try {
            String fileName = name + ".vmdk";
            String dataStorePath = "[" + this.dataStoreName + "]" + this.filePath + "/" + fileName;
            FileBackedVirtualDiskSpec diskSpec = new FileBackedVirtualDiskSpec();
            // XXX check these values and remove hard-coded values
            diskSpec.setCapacityKb(capacityMB * 1024);
            diskSpec.setAdapterType(VirtualDiskAdapterType.lsiLogic.getValue());
            diskSpec.setDiskType(VirtualDiskType.thin.getValue());
            ManagedObjectReference task;
            String status;
            TaskInfo taskInfo = null;
            try {
                VMwareStoragePool.logger.debug("createVirtualDisk name=" + dataStorePath);
                task = connection.getService()
                    .createVirtualDisk_Task(this.virtualDiskManagerRef, dataStorePath, null, diskSpec);
                status = connection.waitForTask(task);
                if (!status.equalsIgnoreCase("success")) {
                    taskInfo = (TaskInfo) connection.getDynamicProperties(task, "info")[0].getVal();
                }
            } catch (Exception ex) {
                VMwareStoragePool.logger.error("Failed to create virtual disk ", ex);
                throw new VMMException("Failed to create volume : " + ex.getMessage());
            }
            if (!status.equalsIgnoreCase("success")) {
                throw new VMMException("Failed to create volume : " + taskInfo.getError().getLocalizedMessage());
            }
            Volume volume = new Volume();
            volume.setStoragePool(this);
            volume.setCapacityMB(capacityMB);
            volume.setAllocationMB(-1);
            volume.setPath(dataStorePath);
            volume.setKey(volume.getPath());
            volume.setFormat(VMwareStoragePool.VMWARE_DISK_FORMAT);
            volume.setName(fileName);
            this.volumeMap.put(volume.getKey(), volume);
            return volume;
        } finally {
            connection.release();
        }
    }

    @Override
    public Volume createVolumeFrom(final String name, final long capacityMB, final String format, final Volume parentVolume,
        final boolean copyOnWrite) throws VMMException {
        // TODO Auto-generated method stub
        return null;
    }

    @Override
    public Volume uploadVolume(final String url, final String format, final String name, final String description,
        final String checkSum) throws InsufficientResourcesException, VMMException {
        // TODO Auto-generated method stub
        return null;
    }

    @Override
    public Map<String, String> getProperties() {
        return this.properties;
    }

}
